#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>

#include "mpsys.h"

#define PES_MAX_LENGTH 100000 /* completely arbitrary */

typedef struct config_st
{
	int dump_pes;
	int dump_ts;
	int dump_es;
	
	int out_pes;
	int out_es;
	
	int filter_audio;
	int filter_video;
	int filter_by_pid;
	int filtered_pid;
	
	int fd_in;
	int fd_out;
} config;

void config_parse (config *cfg, int argc, char **argv)
{
	int i;
	
	assert (cfg);
	
	/* default config */
	cfg->fd_in = 0; /* stdin */
	cfg->fd_out = 1; /* stdout */
	
	cfg->dump_pes = 0;
	cfg->dump_ts = 0;
	cfg->dump_es = 0;
	
	cfg->out_pes = 0;
	cfg->out_es = 0;

	cfg->filter_audio = 0;
	cfg->filter_video = 0;
	cfg->filter_by_pid = 0;
	cfg->filtered_pid = 0;

	
	for (i=1; i<argc; i++)
	{
		char *opt = argv[i];
		
		if (!strcmp (opt, "--dump-pes"))
		{
			cfg->dump_pes = 1;
		}
		else if (!strcmp (opt, "--dump-ts"))
		{
			cfg->dump_ts = 1;
		}
		else if (!strcmp (opt, "--dump-es"))
		{
			cfg->dump_es = 1;
		}
		else if (!strcmp (opt, "--dump"))
		{
			cfg->dump_ts = 1;
			cfg->dump_pes = 1;
			cfg->dump_es = 1;
		}
		else if (!strcmp (opt, "--es"))
		{
			cfg->out_es = 1;
			cfg->out_pes = 0; /* mutually excl. */
		}
		else if (!strcmp (opt, "--pes"))
		{
			cfg->out_pes = 1;
			cfg->out_es = 0; /* mutually excl. */
		}
		else if (!strcmp (opt, "--audio"))
		{
			cfg->filter_audio = 1;
		}
		else if (!strcmp (opt, "--video"))
		{
			cfg->filter_video = 1;
		}
		else if (isdigit((int)opt[0]))
		{ /* assume PID */
			cfg->filter_by_pid = 1;
			cfg->filtered_pid = atoi (opt); 
		}
		else
		{ /* usage */
			fprintf (stderr, "%s [--audio] [--video] [--dump|--dump-pes|--dump-ts] [--es|--pes] [PID]\n", argv[0]);
			fprintf (stderr, "  --dump-ts:  output info on each TS packet on stderr\n");
			fprintf (stderr, "  --dump-pes: output info on each PES packet on stderr\n");
			fprintf (stderr, "  --dump-es:  output info on each ES packet on stderr\n");
			fprintf (stderr, "  --dump:     output info on both PES and TS\n");
			fprintf (stderr, "  --audio:    process only audio stream(s) (PES)\n");
			fprintf (stderr, "  --video:    process only video stream(s) (PES)\n");
			fprintf (stderr, "  --pes:      output raw PES packet on stdout\n");
			fprintf (stderr, "  --es:       output raw ES packet on stdout\n");
			exit (1);
		}
	}
}

/* is this PES packet accepted with current config? */
int filter_pes_accept (pes *p, config *cfg)
{
	assert (pes_invariant (p));
	assert (cfg);
	
	if (cfg->filter_audio && cfg->filter_video)
	{
		return pes_is_stream_audio (p) || pes_is_stream_video (p);
	}
	else if (cfg->filter_audio)
	{
		return pes_is_stream_audio (p);
	}
	else if (cfg->filter_video)
	{
		return pes_is_stream_video (p);
	}
	return 1; /* accept everything by default */
}

/* is this TS packet accepted with current config? */
int filter_ts_accept (ts *t, config *cfg)
{
	assert (ts_invariant (t));
	assert (cfg);
	
	if (cfg->filter_by_pid)
	{
		return (t->pid == cfg->filtered_pid);
	}
	return 1; /* accept everything by default */
}

/* process ES audio */
void process_es_audio (pes *p)
{
	es_audio a;
	buffer_iter data_iter = p->data; /* copy */
	
	assert (pes_is_stream_audio (p));
	
	do{
		buffer b = buffer_iter_buffer_at (&data_iter);
		if (es_audio_parsable (&b))
		{
			es_audio_parse (&a, &b);
			es_audio_print (&a, stderr);
			buffer_iter_skip_bytes (&data_iter, es_audio_frame_length(&a));
		}
		else
		{
			fprintf (stderr, "ES: not an audio ES frame\n");
			buffer_iter_dump (&data_iter, stderr);
			break;
		}
	} while (buffer_iter_available_bytes (&data_iter)>0);
}

/* process incoming PES buffer */
void process_pes (buffer *pes_in, config *cfg)
{
	pes p;
	
	if (pes_parsable (pes_in))
	{	
		pes_parse (&p, pes_in);
		
		if (filter_pes_accept (&p, cfg))
		{
			if (cfg->dump_pes)
			{
				pes_print (&p, stderr);
			}
			if (cfg->dump_es)
			{
				if (pes_is_stream_audio (&p))
				{
					process_es_audio (&p);
				} /* else FIXME: video */
			}
			if (cfg->out_pes)
			{
				write (cfg->fd_out, pes_in->memory, pes_in->len);
			}
			if (cfg->out_es)
			{
				write (cfg->fd_out, 
						buffer_iter_memory_at (&(p.data)), 
						buffer_iter_available_bytes (&(p.data)));
			}
		}
	}
	else
	{
		fprintf (stderr, "Invalid PES packet %s\n", pes_parsable_error (pes_in));
	}
}

static void *malloc_or_exit (int size)
{
	void *result = malloc (size);
	if (!result)
	{
		fprintf (stderr, "Out of memory!");
		exit (2);
	}
	return result;
}
	
ts_payload *ts_payload_find (ts_payload *head, int id)
{
	ts_payload *current = head; 
	while (current && (current->id != id))
	{
		current = current->next;
	}
	assert (!current || (current->id == id));
	return current;
}

ts_payload *ts_payload_add (ts_payload *head, int id)
{
	ts_payload *result;
	buffer result_buffer;
	assert (!ts_payload_find (head, id));
	
	result = malloc_or_exit (sizeof(ts_payload));
	buffer_reset (&result_buffer, malloc_or_exit (PES_MAX_LENGTH), PES_MAX_LENGTH);
	ts_payload_reset (result, id, &result_buffer, head);
	return result;
}

int main (int argc, char **argv)
{
	config cfg;
	
	unsigned char buf_memory[TS_LENGTH];
	buffer buf;

	ts_payload *states = 0;
	
	config_parse (&cfg, argc, argv);

	buffer_reset (&buf, buf_memory, TS_LENGTH);
	
	while (read (cfg.fd_in, buf.memory, TS_LENGTH) == TS_LENGTH)
	{
		if (ts_parsable (&buf))
		{
			ts packet;
			ts_parse (&packet, &buf);
			
			if (filter_ts_accept (&packet, &cfg))
			{
				ts_payload *state;
				
				if (cfg.dump_ts)
				{
					ts_print (&packet, stderr);
				}

				state = ts_payload_find (states, packet.pid);
				if (!state)
				{ /* create state if none for this PID */
					if (ts_starts_payload (&packet))
					{
						states = ts_payload_add (states, packet.pid);
						state = ts_payload_find (states, packet.pid);
					}
					else
					{
						fprintf (stderr, "Ignoring incomplete trailer for %d\n", packet.pid);
					}
				}
				
				if (state)
				{
					/* is this the first TS of a payload? */
					if (ts_starts_payload (&packet))
					{
						/* process previous PES if any */
						if (state->payload_writer.byte_pos > 0)
						{
							/* truncate to real TS data section */
							buffer pes_in;
							buffer_reset (&pes_in, state->payload_writer.memory, state->payload_writer.byte_pos);
							process_pes (&pes_in, &cfg);
						}
						/* resets write */
						ts_payload_reset_payload_writer (state);
					}
					
					if (ts_payload_ts_applies (state, &packet))
					{
						ts_payload_ts_apply (state, &packet);
					}
					else
					{
						fprintf (stderr, "TS packet does not apply: %s\n", 
								ts_payload_ts_applies_error (state, &packet));
					}
				}
			} /* filter ts */
		}
		else
		{
			fprintf (stderr, "Not a TS packet: %s\n", ts_parsable_error (&buf));
		}
	}
	
	fprintf (stderr, "No more packets.\n");
	
	return 0;
}
